@@ -17,6 +17,7 @@ import Network.QUIC.Connector
 import Network.QUIC.Crypto
 import Network.QUIC.Exception
 import Network.QUIC.Imports
+import Network.QUIC.Info
 import Network.QUIC.Logger
 import Network.QUIC.Packet
 import Network.QUIC.Parameters
@@ -255,7 +256,7 @@ processFrame conn lvl (StopSending sid err) = do
         Nothing -> do
             when (isInitiated conn sid) $
                 closeConnection StreamStateError "No such stream for STOP_SENDING"
-        Just _strm -> sendFrames conn lvl [ResetStream sid err 0]
+        Just _strm -> sendFramesLim conn lvl [ResetStream sid err 0]
 processFrame _ _ (CryptoF _ "") = return ()
 processFrame conn lvl (CryptoF off cdat) = do
     when (lvl == RTT0Level) $
@@ -343,9 +344,11 @@ processFrame conn lvl (MaxStreamData sid n) = do
         closeConnection StreamStateError "Receive-only stream"
     mstrm <- findStream conn sid
     case mstrm of
-        Nothing -> do
-            when (isInitiated conn sid) $
-                closeConnection StreamStateError "No such stream for MAX_STREAM_DATA"
+        Nothing ->
+            when (isInitiated conn sid) $ do
+                curSid <- getMyStreamId conn
+                when (sid > curSid) $
+                    closeConnection StreamStateError "No such stream for MAX_STREAM_DATA"
         Just strm -> setTxMaxStreamData strm n
 processFrame conn lvl (MaxStreams dir n) = do
     when (lvl == InitialLevel || lvl == HandshakeLevel) $
@@ -365,13 +368,15 @@ processFrame _conn lvl (StreamsBlocked _dir n) = do
 processFrame conn lvl (NewConnectionID cidInfo rpt) = do
     when (lvl == InitialLevel || lvl == HandshakeLevel) $
         closeConnection ProtocolViolation "NEW_CONNECTION_ID in Initial or Handshake"
-    addPeerCID conn cidInfo
+    ok <- addPeerCID conn cidInfo
+    unless ok $
+        closeConnection ConnectionIdLimitError "NEW_CONNECTION_ID limit error"
     let (_, cidlen) = unpackCID $ cidInfoCID cidInfo
     when (cidlen < 1 || 20 < cidlen || rpt > cidInfoSeq cidInfo) $
         closeConnection FrameEncodingError "NEW_CONNECTION_ID parameter error"
     when (rpt >= 1) $ do
         seqNums <- setPeerCIDAndRetireCIDs conn rpt
-        sendFrames conn RTT1Level $ map RetireConnectionID seqNums
+        sendFramesLim conn RTT1Level $ map RetireConnectionID seqNums
 processFrame conn RTT1Level (RetireConnectionID sn) = do
     mcidInfo <- retireMyCID conn sn
     case mcidInfo of
@@ -381,7 +386,7 @@ processFrame conn RTT1Level (RetireConnectionID sn) = do
                 unregister <- getUnregister conn
                 unregister cid
 processFrame conn RTT1Level (PathChallenge dat) =
-    sendFrames conn RTT1Level [PathResponse dat]
+    sendFramesLim conn RTT1Level [PathResponse dat]
 processFrame conn RTT1Level (PathResponse dat) =
     -- RTT0Level falls intentionally
     checkResponse conn dat
@@ -407,6 +412,7 @@ processFrame conn lvl HandshakeDone = do
         clearCryptoStream conn HandshakeLevel
         clearCryptoStream conn RTT1Level
     setConnectionEstablished conn
+    getConnectionInfo conn >>= onConnectionEstablished (connHooks conn)
     -- to receive NewSessionTicket
     fire conn (Microseconds 1000000) $ killHandshaker conn lvl
 processFrame _ _ _ = closeConnection ProtocolViolation "Frame is not allowed"
